import { DescribeImagesCommand, EC2Client } from '@aws-sdk/client-ec2'
import { spawn } from 'child_process'
import { mkdtemp, writeFile } from 'fs/promises'
import { tmpdir } from 'os'
import { join } from 'path'
import { VMSpec } from '../../Driver'
import { getAuxVmImageName } from './getAuxVmImageName'
import { getImageNameGenerationData } from './getImageNameGenerationData'
import { getPackerTemplate } from './getPackerTemplate'

export async function buildAuxVmImage(
  taskFamilyDirectory: string,
  vmSpec: VMSpec,
  onBuildOutput: (type: 'stdout' | 'stderr', chunk: string) => void,
): Promise<string> {
  const imageNameGenerationData = await getImageNameGenerationData(vmSpec)

  if (!vmSpec.build_steps || vmSpec.build_steps.length === 0) {
    return imageNameGenerationData.sourceAmiId
  }

  const imageName = await getAuxVmImageName(taskFamilyDirectory, vmSpec)

  const ec2Client = new EC2Client()
  const describeImagesCommand = new DescribeImagesCommand({ Filters: [{ Name: 'name', Values: [imageName] }] })
  let images = await ec2Client.send(describeImagesCommand)
  if (images.Images && images.Images.length >= 1) {
    const image = images.Images[0]
    return image.ImageId!
  }

  const temporaryDirectory = await mkdtemp(join(tmpdir(), 'metr-task-standard-packer-build-'))

  const packerVariables = {
    build_instance_type: imageNameGenerationData.buildInstanceType,
    region: imageNameGenerationData.region,
    source_ami_id: imageNameGenerationData.sourceAmiId,
    ssh_username: imageNameGenerationData.sshUsername,
    ami_name: imageName,
  }
  const packerVariablesJsonPath = join(temporaryDirectory, 'metr-task-standard.pkrvars.json')
  await writeFile(packerVariablesJsonPath, JSON.stringify(packerVariables, /* replacer */ null, 2) + '\n')

  const packerTemplatePath = join(temporaryDirectory, 'aux-vm.pkr.hcl')
  await writeFile(packerTemplatePath, await getPackerTemplate(taskFamilyDirectory, imageNameGenerationData.buildSteps))

  const packer = spawn('packer', ['build', '-color=false', '-var-file', packerVariablesJsonPath, packerTemplatePath], {
    stdio: 'pipe',
    cwd: taskFamilyDirectory,
  })

  packer.stdout.on('data', chunk => {
    onBuildOutput('stdout', chunk.toString())
  })
  packer.stderr.on('data', chunk => {
    onBuildOutput('stderr', chunk.toString())
  })

  const exitCode = await new Promise<number>((resolve, reject) => {
    packer.on('error', error => {
      reject(new Error(`Packer build failed to start: ${error.message}`))
    })

    packer.on('close', resolve)
  })
  if (exitCode !== 0) {
    throw new Error(`Packer build failed with exit code ${exitCode}.`)
  }

  images = await ec2Client.send(describeImagesCommand)
  if (images.Images && images.Images.length >= 1) {
    const image = images.Images[0]
    return image.ImageId!
  }

  throw new Error(`AMI build succeeded, but no AMI with name ${imageName} could be found afterwards.`)
}
