import { BuildStep } from '../../Driver'
import { dedent } from '../../lib/dedent'
import { validateBuildSteps } from './validateBuildSteps'

// Packer templates aren't flexible enough to support templating based on an array of build steps of different types.
// We still use user variables for the other fields so that we can mostly rely on Packer's user variable escaping logic.
export async function getPackerTemplate(taskFamilyDirectory: string, buildStepsBeforeValidation: BuildStep[]) {
  const validatedBuildSteps = await validateBuildSteps(taskFamilyDirectory, buildStepsBeforeValidation)

  return dedent`
    variable "ami_name" {
      type        = string
      description = "The name of the AMI that Packer will create"
    }

    variable "build_instance_type" {
      type        = string
      description = "The instance type to use when building the AMI"
    }

    variable "region" {
      type        = string
      description = "The region to build the AMI in"
    }

    variable "source_ami_id" {
      type        = string
      description = "The ID of the source AMI to base the new AMI on"
    }

    variable "ssh_username" {
      type        = string
      description = "The username to use when connecting to the build instance"
    }

    source "amazon-ebs" "aux-vm" {
      ami_name      = var.ami_name
      instance_type = var.build_instance_type
      source_ami    = var.source_ami_id
      region        = var.region
      ssh_username  = var.ssh_username
    }

    build {
      sources = ["source.amazon-ebs.aux-vm"]

      ${validatedBuildSteps
        .map(validatedBuildStep => {
          // getPackerTemplate must take care to escape everything it adds to the Packer template from buildSteps. Otherwise, a task author
          // could write a task that executes arbitrary code on the computer running this template, by injecting a local shell provisioner
          // into the template.
          // The simplest way to do this is to wrap everything added to the template in JSON.stringify.
          if (validatedBuildStep.type === 'file') {
            return `provisioner "file" {
              source      = ${JSON.stringify(validatedBuildStep.sourceWithinTaskFamilyDirectory)}
              destination = ${JSON.stringify(validatedBuildStep.destination)}
            }`
          } else {
            return `provisioner "shell" {
              inline = ${JSON.stringify(validatedBuildStep.commands)}
            }`
          }
        })
        .join('\n\n')}
    }
  `
}
